import 'dart:ffi';
import 'package:ffi/ffi.dart';
import 'package:ffi_helper/src/array_utils.dart';
import 'package:meta/meta.dart';

import 'array.dart';

/// Instead of doing many little allocations, [Arena]
/// allocates a large chunk of memory which can be sliced into multiple [ArrayView]s.
/// This can significantly improve performance and it makes
/// memory management simpler.
/// ````
/// final arena = Uint8Arena.allocate(10);
/// final array = arena.newArray(5);
/// final array2 = arena.newArray(5);
/// void some_native_c_api(array.ptr, array.view.length, array2.ptr, array2.view.length);
/// arena.delete();
/// ````
class Arena<T extends NativeType> {
  int _index = 0;

  /// Total length of allocated memory
  final int length;

  /// [Pointer] to allocated buffer
  @protected
  final Pointer<T> buffer;
  final GetView<T> _getView;

  /// Left free space of Arena.
  int get freeSpace => length - _index;

  Arena(this.length, this._getView) : buffer = allocate(count: length);

  Arena.fromPtr(this.buffer, this.length, this._getView);

  /// Returns a slice of [length] elements from the pre-allocated array of Arena.
  /// When [length] exceeds [freeSpace] it throws a [RangeError].
  @mustCallSuper
  ArrayView<T> newArray(int length) {
    assert(length > 0);
    if (freeSpace < length) {
      throw RangeError('Not enough free space');
    }
    final ptrAtIndex = buffer.elementAt(_index);
    final view = _getView(ptrAtIndex, length);
    final arr = ArrayView<T>.fromPtr(ptrAtIndex, view);
    _index += length;
    return arr;
  }

  /// Deletes all allocated memory.
  void delete() {
    free(buffer);
  }
}

extension Uint8Arena on Arena<Uint8> {
  static Arena<Uint8> allocate(int length) => Arena(length, getUint8View);
  static Arena<Uint8> fromPtr(Pointer<Uint8> ptr, int length) =>
      Arena.fromPtr(ptr, length, getUint8View);
}
