import 'dart:ffi';
import 'package:ffi/ffi.dart';
import 'package:meta/meta.dart';
import 'dart:typed_data';
import 'array_utils.dart';

typedef GetView<T extends NativeType> = List<int> Function(
    Pointer<T> ptr, int length);

/// Thin wrapper around a [Pointer] and its associated typed list.
/// It is meant to be used with extensions like [Uint8Array].
/// It can be extended to implement a different allocator or deallocator.
class Array<T extends NativeType> {
  @doNotStore
  final Pointer<T> ptr;

  final List<int> _view;

  @doNotStore
  List<int> get untypedView => _view;

  Array.fromPtr(this.ptr, this._view);

  Array._(this.ptr, int length, GetView<T> getView)
      : _view = getView(ptr, length);

  Array(int length, GetView<T> getView)
      : this._(allocate<T>(count: length), length, getView);

  Array._fromList(this.ptr, Iterable<int> list, GetView<T> getView)
      : _view = getView(ptr, list.length)..setAll(0, list);
  Array.fromList(Iterable<int> list, GetView<T> getView)
      : this._fromList(allocate<T>(count: list.length), list, getView);

  void delete() {
    free(ptr);
  }

  @override
  bool operator ==(dynamic arr) {
    if (arr is Pointer) {
      return arr == ptr;
    }
    if (arr is Array) {
      return arr.ptr == ptr;
    }
    return false;
  }

  @override
  int get hashCode => ptr.hashCode;

  @override
  String toString() {
    return '${runtimeType}($ptr, $_view)';
  }
}

extension Uint8Array on Array<Uint8> {
  Uint8List get view => untypedView;
  static Array<Uint8> fromList(Iterable<int> list) =>
      Array.fromList(list, getUint8View);

  static Array<Uint8> allocate(int length) => Array(length, getUint8View);
}

/// Wrapper for an [Array] with an empty [delete] method.
class ArrayView<T extends NativeType> extends Array<T> {
  ArrayView.fromPtr(Pointer<T> ptr, List<int> view) : super.fromPtr(ptr, view);

  /// Doesn't do anything. It's only there to be compatible with [Array].
  @override
  void delete() {}
}
