import 'package:meta/meta.dart';

import 'array.dart';
import 'dart:ffi';

/// Instead of allocating many arrays, share one array which
/// gets deleted when all references to it got deleted. This can lead to
/// better performance and reduced memory usage.
///
/// Gets lazily initialized when the first reference gets created.
/// It is safe to use even after deletion; it just reallocates new memory for new references.
/// Not deleting all references causes memory leaks.
/// ````
/// final shared = SharedUint8Array<Uint8>(10);
/// // All instances of [WritesToArray] share the same [SharedArray] so they don't have to all
/// // allocate their own array.
/// class WritesToArray {
///   final _buffer = shared.newRef();
///
///   Uint8List someNativeWrapper(){
///     writeToBuffer(_buffer.ptr);
///     return Uint8List.fromList(_buffer.view);
///   }
///   void delete() => _buffer.delete();
/// }
/// ````
class SharedArray<T extends NativeType> {
  int _refCount = 0;
  int get refCount => _refCount;
  Array<T> _array;
  final Array<T> Function(int) _allocateArray;
  final int length;

  /// [_allocateArray] allocates an [Array] of [length] elements
  SharedArray(this.length, this._allocateArray);

  /// Create new reference to this SharedArray.
  @mustCallSuper
  ArrayRef<T> newRef() {
    if (_refCount == 0) {
      _array = _allocateArray(length);
    }
    _refCount++;
    return ArrayRef._(this);
  }
}

extension SharedUint8Array on SharedArray<Uint8> {
  static SharedArray<Uint8> allocate(int length) =>
      SharedArray(length, Uint8Array.allocate);
}

/// Reference to a [SharedArray].
class ArrayRef<T extends NativeType> implements Array<T> {
  final SharedArray<T> _shared;
  @override
  @doNotStore
  List<int> get untypedView => _shared._array.untypedView;
  @override
  @doNotStore
  Pointer<T> get ptr => _shared._array.ptr;
  bool _freed = false;

  ArrayRef._(this._shared);
  @override
  @mustCallSuper
  void delete() {
    if (_freed) {
      return;
    }
    assert(_shared._refCount > 0);
    _shared._refCount--;
    if (_shared._refCount == 0) {
      _shared._array.delete();
    }
    _freed = true;
  }

  @override
  String toString() {
    return _shared._array.toString();
  }
}
